"""
A custom gunicorn gevent worker for Flask applications.

This worker is designed to handle SIGINT and SIGTERM gracefully, by
closing all client connections and logging the stacktrace before
exiting.

## Usage
Run gunicorn in the usual way, but specify the worker class as LogWorker.

talisker.gunicorn.gevent webapp.app:app \
    -k canonicalwebteam.flask_base.worker.LogWorker

"""

from __future__ import annotations

import logging
import os
import secrets
import traceback
from typing import TYPE_CHECKING

from gunicorn.workers.ggevent import GeventWorker

if TYPE_CHECKING:
    from socket import socket
    from types import FrameType


logger = logging.getLogger("gunicorn.error")


class LogWorker(GeventWorker):
    def __init__(self, *args: tuple, **kwargs: dict) -> None:
        super().__init__(*args, **kwargs)
        self.instance_id = secrets.token_hex(6)
        self.clients: list[socket] = []

    def _log(self, msg: str) -> None:
        msg = f"[LOG WORKER][{self.instance_id}]: {msg}"
        self.log.info(msg)

    def close_clients_gracefully(self) -> None:
        """Send a 200 response to all registered clients"""
        self._log(f"closing {len(self.clients)} clients gracefully")
        msg = "OK"
        size = len(msg)
        response_data = (
            f"HTTP/1.1 200 {msg}\r\nContent-Length: {size}\r\n\r\n{msg}"
        )
        response_data = response_data.encode("utf-8")
        try:
            for conn in self.clients:
                conn.sendall(response_data)
                conn.close()
        except Exception as e:  # noqa: BLE001
            self._log("Unable to close client: " + str(e))

    def notify_error(self, sig: int) -> None:
        """Print recent traceback logs."""
        self._log(f"notifying errors with signal {sig}")
        self._log("Stacktrace: " + traceback.format_exc())

    def handle(self, listener: socket, client: socket, addr: tuple) -> None:
        """Handle a new client connection."""
        # Register client connections to this worker
        self.clients.append(listener)
        super().handle(listener, client, addr)

    def handle_termination_gracefully(
        self,
        sig: int,
        frame: FrameType | None,
    ) -> None:
        """Handle termination signal gracefully"""
        self._log(f"handling signal {sig}")
        self.notify_error(sig)
        self.close_clients_gracefully()
        self._log(f"closing worker {self.instance_id}")
        os._exit(0)  # exit immediately, avoiding later exception catches

    def handle_exit(self, sig: int, frame: FrameType | None) -> None:
        """Handle SIGTERM gracefully"""
        self.handle_termination_gracefully(sig, frame)

    def handle_quit(self, sig: int, frame: FrameType | None) -> None:
        """Handle SIGQUIT gracefully"""
        self.handle_termination_gracefully(sig, frame)

    def handle_interrupt(self, sig: int, frame: FrameType | None) -> None:
        """Handle SIGINT gracefully"""
        self.handle_termination_gracefully(sig, frame)



"""
This module provides a middleware that adjusts the WSGI environ based on
``X-Forwarded-`` headers that proxies in front of an application may
set.

When an application is running behind a proxy server, WSGI may see the
request as coming from that server rather than the real client. Proxies
set various headers to track where the request actually came from.

This is based on werkzeug's ProxyFix middleware:
https://github.com/pallets/werkzeug/blob/main/src/werkzeug/middleware/proxy_fix.py

With additional support for `X-Original-Forwarded-For` header.
"""

from __future__ import annotations

import typing as t

from werkzeug.http import parse_list_header


class ProxyFix:
    """Adjust the WSGI environ based on ``X-Forwarded-`` that proxies in
        front of the application may set.

        -   ``X-Forwarded-For`` sets ``REMOTE_ADDR``.
        -   ``X-Original-Forwarded-For`` sets ``REMOTE_ADDR``(and takes precedence over ``X-Forwarded-For`` if configured). # noqa
        -   ``X-Forwarded-Proto`` sets ``wsgi.url_scheme``.
        -   ``X-Forwarded-Host`` sets ``HTTP_HOST``, ``SERVER_NAME``
     and ``SERVER_PORT``.
        -   ``X-Forwarded-Port`` sets ``HTTP_HOST`` and ``SERVER_PORT``.
        -   ``X-Forwarded-Prefix`` sets ``SCRIPT_NAME``.

        You must tell the middleware how many proxies set each header so it
        knows what values to trust. It is a security issue to trust values
        that came from the client rather than a proxy.

        The original values of the headers are stored in the WSGI
        environ as ``werkzeug.proxy_fix.orig``, a dict.

        :param app: The WSGI application to wrap.
        :param x_for: Number of values to trust for ``X-Forwarded-For``.
        :param x_original_for: Number of values to trust for
    ``X-Original-Forwarded-For``.
        :param x_proto: Number of values to trust for ``X-Forwarded-Proto``.
        :param x_host: Number of values to trust for ``X-Forwarded-Host``.
        :param x_port: Number of values to trust for ``X-Forwarded-Port``.
        :param x_prefix: Number of values to trust for ``X-Forwarded-Prefix``.

        ```
        from werkzeug.middleware.proxy_fix import ProxyFix
        # App is behind one proxy that sets the -For and -Host headers.
        app = ProxyFix(app, x_for=1, x_host=1)
        ```
    """

    def __init__(
        self,
        app,
        x_for: int = 1,
        x_original_for: int = 0,
        x_proto: int = 1,
        x_host: int = 0,
        x_port: int = 0,
        x_prefix: int = 0,
    ) -> None:
        self.app = app
        self.x_for = x_for
        self.x_original_for = x_original_for
        self.x_proto = x_proto
        self.x_host = x_host
        self.x_port = x_port
        self.x_prefix = x_prefix

    def _get_real_value(self, trusted: int, value: str | None) -> str | None:
        """Get the real value from a list header based on the configured
        number of trusted proxies.

        :param trusted: Number of values to trust in the header.
        :param value: Comma separated list header value to parse.
        :return: The real value, or ``None`` if there are fewer values
            than the number of trusted proxies.
        """
        if not (trusted and value):
            return None
        values = parse_list_header(value)
        if len(values) >= trusted:
            return values[-trusted]
        return None

    def __call__(self, environ, start_response) -> t.Iterable[bytes]:
        """Modify the WSGI environ based on the various forwarding headers
        before calling the wrapped application. Store the original
        environ values in ``werkzeug.proxy_fix.orig``.
        """
        environ_get = environ.get
        orig_remote_addr = environ_get("REMOTE_ADDR")
        orig_wsgi_url_scheme = environ_get("wsgi.url_scheme")
        orig_http_host = environ_get("HTTP_HOST")
        environ.update(
            {
                "werkzeug.proxy_fix.orig": {
                    "REMOTE_ADDR": orig_remote_addr,
                    "wsgi.url_scheme": orig_wsgi_url_scheme,
                    "HTTP_HOST": orig_http_host,
                    "SERVER_NAME": environ_get("SERVER_NAME"),
                    "SERVER_PORT": environ_get("SERVER_PORT"),
                    "SCRIPT_NAME": environ_get("SCRIPT_NAME"),
                }
            }
        )

        # Process X-Forwarded-For
        x_for = self._get_real_value(
            self.x_for, environ_get("HTTP_X_FORWARDED_FOR")
        )
        if x_for:
            environ["REMOTE_ADDR"] = x_for

        # Process X-Original-Forwarded-For.
        # If present, this header takes precedence over X-Forwarded-For.
        x_original_for = self._get_real_value(
            self.x_original_for, environ_get("HTTP_X_ORIGINAL_FORWARDED_FOR")
        )
        if x_original_for:
            environ["REMOTE_ADDR"] = x_original_for

        x_proto = self._get_real_value(
            self.x_proto, environ_get("HTTP_X_FORWARDED_PROTO")
        )
        if x_proto:
            environ["wsgi.url_scheme"] = x_proto

        x_host = self._get_real_value(
            self.x_host, environ_get("HTTP_X_FORWARDED_HOST")
        )
        if x_host:
            environ["HTTP_HOST"] = environ["SERVER_NAME"] = x_host
            # If the host contains a port
            # (and isn't an IPv6 literal without a port),
            # split it out.
            if ":" in x_host and not x_host.endswith("]"):
                environ["SERVER_NAME"], environ["SERVER_PORT"] = x_host.rsplit(
                    ":", 1
                )

        x_port = self._get_real_value(
            self.x_port, environ_get("HTTP_X_FORWARDED_PORT")
        )
        if x_port:
            host = environ.get("HTTP_HOST")
            if host:
                # If the host contains a port
                # (and isn't an IPv6 literal without a port),
                # remove the existing port before appending the new one.
                if ":" in host and not host.endswith("]"):
                    host = host.rsplit(":", 1)[0]
                environ["HTTP_HOST"] = f"{host}:{x_port}"
            environ["SERVER_PORT"] = x_port

        x_prefix = self._get_real_value(
            self.x_prefix, environ_get("HTTP_X_FORWARDED_PREFIX")
        )
        if x_prefix:
            environ["SCRIPT_NAME"] = x_prefix

        return self.app(environ, start_response)



#! /usr/bin/env python3

from setuptools import find_packages, setup

setup(
    name="canonicalwebteam.flask-base",
    version="2.5.0",
    description=(
        "Flask extension that applies common configurations"
        "to all of webteam's flask apps."
    ),
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Canonical webteam",
    author_email="webteam@canonical.com",
    url="https://github.com/canonical-web-and-design/canonicalwebteam.flask-base",
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Topic :: Internet :: WWW/HTTP :: WSGI",
        "Topic :: Internet :: WWW/HTTP :: WSGI :: Application",
        "Topic :: Internet :: WWW/HTTP :: WSGI :: Middleware",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.12",
    ],
    python_requires=">=3.10",
    packages=find_packages(),
    install_requires=[
        "canonicalwebteam.yaml-responses[flask] (>=1,<2)",
        "talisker[gunicorn,gevent,flask,prometheus,raven] >= 0.21.4",
        # TODO: Remove this once Flask is updated to 3.x.x
        "Werkzeug < 3.0.0",
        # Use latest version of Flask once Talisker supports werkzeug >=3
        "flask==2.3.3",
        "flask-compress==1.17",
    ],
    dependency_links=[],
    include_package_data=True,
    project_urls={},
)
